"""Handle file operations"""

import hashlib
import json
import logging
import os.path
from pathlib import Path
import tkinter as tk
from tkinter import filedialog, messagebox, simpledialog
from typing import Any, Callable, Final, TypedDict, Literal, Optional

import regex as re

from guiguts.mainwindow import mainimage
from guiguts.maintext import (
    maintext,
    PAGE_FLAG_TAG,
    PAGEMARK_PIN,
    BOOKMARK_TAG,
    img_from_page_mark,
    page_mark_from_img,
)
from guiguts.page_details import (
    PageDetail,
    PageDetails,
    PAGE_LABEL_PREFIX,
    NUMBER_INCREMENT,
    STYLE_ARABIC,
    STYLE_DITTO,
)
from guiguts.preferences import preferences, PrefKey
from guiguts.project_dict import ProjectDict, GOOD_WORDS_FILENAME, BAD_WORDS_FILENAME
from guiguts.root import root

from guiguts.utilities import (
    load_dict_from_json,
    IndexRowCol,
    IndexRange,
    sound_bell,
    folder_dir_str,
)
from guiguts.widgets import grab_focus, ToplevelDialog

logger = logging.getLogger(__package__)

NUM_RECENT_FILES = 9
BINFILE_SUFFIX = ".json"

BINFILE_KEY_MD5CHECKSUM: Final = "md5checksum"
BINFILE_KEY_PAGEDETAILS: Final = "pagedetails"
BINFILE_KEY_INSERTPOS: Final = "insertpos"
BINFILE_KEY_INSERTPOSPEER: Final = "insertpospeer"
BINFILE_KEY_IMAGEDIR: Final = "imagedir"
BINFILE_KEY_PROJECTID: Final = "projectid"
BINFILE_KEY_LANGUAGES: Final = "languages"
BINFILE_KEY_BOOKMARKS: Final = "bookmarks"

PAGE_FLAGS_NONE = 0
PAGE_FLAGS_SOME = 1
PAGE_FLAGS_ALL = 2
PAGE_FLAG_PREFIX = "Img"
PAGE_FLAG_START = "["
PAGE_FLAG_START_E = re.escape(PAGE_FLAG_START)
PAGE_FLAG_END = "]"
PAGE_FLAG_END_E = re.escape(PAGE_FLAG_END)
PAGE_FLAG_SEP = "|"
PAGE_FLAG_SEP_E = re.escape(PAGE_FLAG_SEP)
PAGE_FLAG_ARABIC = "A"
PAGE_FLAG_ROMAN = "R"
# 3 capture groups for img, style (Roman/Arabic), number (or +1/None)
# Label is calculated from other info after file load.
PAGE_FLAG_REGEX = rf"{PAGE_FLAG_START_E}{PAGE_FLAG_PREFIX}(.+?){PAGE_FLAG_SEP_E}(.+?){PAGE_FLAG_SEP_E}(.+?){PAGE_FLAG_END_E}"

PAGE_SEPARATOR_REGEX = r"File:.+?([^/\\ ]+)\.(png|jpg)"

BOOKMARK_BASE = "Bookmark"


class BinDict(TypedDict):
    """Dictionary for saving to bin file."""

    md5checksum: str
    pagedetails: PageDetails
    insertpos: str
    insertpospeer: str
    imagedir: str
    projectid: str
    languages: str
    bookmarks: dict[str, str]


class File:
    """Handle data and actions relating to the main text file.

    Attributes:
        _filename: Current filename.
        _filename_callback: Function to be called whenever filename is set.
        _languages: Languages used in file.
        _languages_callback: Function to be called whenever languages are set.
        _image_dir: Directory containing scan images.
        _project_id: Project ID.
    """

    def __init__(
        self,
        filename_callback: Callable[[], None],
        languages_callback: Callable[[], None],
    ):
        """
        Args:
            filename_callback: Function to be called whenever filename is set.
        """
        self._filename = ""
        self._filename_callback = filename_callback
        self._image_dir = ""
        self._project_id = ""
        self._languages = "en"
        self._languages_callback = languages_callback
        self.page_details = PageDetails()
        self.project_dict = ProjectDict()

    @property
    def filename(self) -> str:
        """Name of currently loaded file.

        When assigned to, executes callback function to update interface"""
        return self._filename

    @filename.setter
    def filename(self, value: str) -> None:
        self._filename = value
        self._filename_callback()

    @property
    def image_dir(self) -> Any:
        """Directory containing scan images.

        If unset, defaults to pngs subdir of project dir"""
        if (not self._image_dir) and self.filename:
            proj_dir = os.path.dirname(self.filename)
            maybe_dir = os.path.join(proj_dir, "pngs")
            if (not os.path.isdir(maybe_dir)) and self.project_id:
                maybe_dir = os.path.join(proj_dir, self.project_id + "_images")
            if os.path.isdir(maybe_dir):
                self._image_dir = maybe_dir
        return self._image_dir

    @image_dir.setter
    def image_dir(self, value: str) -> None:
        self._image_dir = value

    @property
    def project_id(self) -> Any:
        """ID of project, e.g. "projectID0123456789abc".

        If unset, extract from project comments filename.
        """
        if (not self._project_id) and self.filename:
            proj_dir = os.path.dirname(self.filename)
            pathlist = Path(proj_dir).glob("projectID*_comments.html")
            for path in pathlist:
                path_str = str(path)
                if match := re.search(r"(projectID[0-9a-f]*)_comments\.html", path_str):
                    self._project_id = match[1]
                    break
        return self._project_id

    @project_id.setter
    def project_id(self, value: str) -> None:
        self._project_id = value

    @property
    def languages(self) -> Any:
        """Languages of currently loaded file.

        Multiple languages are separated by "+"
        When assigned to, executes callback function to update interface"""
        return self._languages

    @languages.setter
    def languages(self, value: str) -> None:
        value = value.strip()
        if not value:
            value = self._languages if self._languages else "en"
        if re.fullmatch(r"[a-z_]+(\+[a-z_]+)*", value) is None:
            logger.error("Invalid language code(s): valid examples are 'en' or 'de fr'")
            return
        self._languages = value
        self._languages_callback()
        preferences.set(PrefKey.DEFAULT_LANGUAGES, self._languages)
        # Inform maintext, so text manipulation algorithms there can check languages
        maintext().set_languages(self._languages)

    def reset(self) -> None:
        """Reset file internals to defaults, e.g. filename, page markers, etc.

        Also close any open dialogs, since they will refer to the previous file."""
        self.filename = ""
        self.image_dir = ""
        self.project_id = ""
        self.remove_page_marks()
        self.page_details = PageDetails()
        ToplevelDialog.close_all()

    def open_file(self, filename: str = "") -> str:
        """Open and load a text file.

        Args:
            Optional filename - if not given, ask user to select file
        Returns:
            Name of file opened - empty string if cancelled.
        """
        if self.check_save():
            if not filename:
                filename = filedialog.askopenfilename(
                    filetypes=(
                        ("Text files", "*.txt *.html *.htm"),
                        ("All files", "*.*"),
                    )
                )
            if filename:
                self.load_file(filename)
            grab_focus(root(), maintext())
        return filename

    def close_file(self) -> None:
        """Close current file, leaving an empty file."""
        if self.check_save():
            self.reset()
            maintext().do_close()

    def load_file(self, filename: str) -> None:
        """Load file & bin file.

        Args:
            filename: Name of file to be loaded. Bin filename has ".bin" appended.
        """
        self.reset()
        try:
            maintext().do_open(filename)
        except FileNotFoundError:
            logger.error(f"Unable to open {filename}")
            self.remove_recent_file(filename)
            self.filename = ""
            return
        maintext().set_insert_index(IndexRowCol(1, 0))
        self.languages = preferences.get(PrefKey.DEFAULT_LANGUAGES)
        bin_matches_file = self.load_bin(filename)
        self.mark_page_boundaries()
        flags_found = (
            self.update_page_marks_from_flags() or self.update_page_marks_from_ppgen()
        )
        if not bin_matches_file:
            # Inform user that bin doesn't match
            # If flags are present, user has used page marker flags to store the
            # page boundary positions while using other editor, so only issue a
            # warning. Otherwise, other failure, so it's an error.
            msg_start = (
                "Main file and bin (.json) file do not match.\n"
                "  File may have been edited in a different editor.\n"
            )
            if flags_found:
                logger.warning(
                    msg_start + "  However, page marker flags were detected,\n"
                    "  so page boundary positions were updated if necessary."
                )
            else:
                logger.error(
                    msg_start + "  You may continue, but page boundary positions\n"
                    "  may not be accurate."
                )

        self.page_details.recalculate()
        self.store_recent_file(filename)
        self.project_dict.load(filename)
        # Load complete, so set filename (including side effects)
        self.filename = filename

    def save_file(self) -> str:
        """Save the current file.

        Returns:
            Current filename or None if save is cancelled
        """
        if self.filename:
            maintext().do_save(self.filename)
            self.save_bin(self.filename)
            return self.filename
        return self.save_as_file()

    def save_as_file(self) -> str:
        """Save current text as new file.

        Returns:
            Chosen filename or None if save is cancelled
        """
        # If no current extension, or ".txt" extension, set extension to ".html"
        # if it's an HTML file. Similarly for text files, otherwise leave alone.
        cur_ext = os.path.splitext(self.filename)[1]
        html_file = maintext().get("1.0", "1.end") == "<!DOCTYPE html>"
        if html_file and cur_ext in ("", ".txt", ".html"):
            extension = ".html"
        elif not html_file and cur_ext in ("", ".txt"):
            extension = ".txt"
        else:
            extension = cur_ext
        match extension:
            case ".html":
                file_type = "HTML files"
            case ".txt":
                file_type = "Text files"
            case _:
                file_type = f"{extension} files"
        suggested_fn = os.path.splitext(self.filename)[0] + extension

        initialfile = (
            os.path.basename(suggested_fn) if self.filename else f"untitled{extension}"
        )
        if fn := filedialog.asksaveasfilename(
            initialfile=initialfile,
            initialdir=os.path.dirname(suggested_fn),
            filetypes=[(file_type, f"*{extension}"), ("All files", "*")],
            title="Save As",
        ):
            self.store_recent_file(fn)
            self.filename = fn
            self.save_file()
        grab_focus(root(), maintext())
        return fn

    def save_copy_as_file(self) -> None:
        """Save copy of current text as new file, without affecting
        current filename or "modified" flag."""
        if fn := filedialog.asksaveasfilename(
            initialfile=os.path.basename(self.filename),
            initialdir=os.path.dirname(self.filename),
            filetypes=[("All files", "*")],
            title="Save a Copy As",
        ):
            self.save_copy(fn)
        grab_focus(root(), maintext())

    def save_copy(self, filename: str) -> None:
        """Save copy of current text as new file, without affecting
        current filename or "modified" flag.

        Args:
            filename: Name to save file as.
        """
        self.store_recent_file(filename)
        self.store_recent_file(self.filename)
        maintext().do_save(filename, clear_modified_flag=False)
        self.save_bin(filename)

    def check_save(self) -> bool:
        """If file has been edited, check if user wants to save,
        or discard, or cancel the intended operation.

        Returns:
            True if OK to continue with intended operation.
        """
        # Good(?) place to ensure image position gets saved
        mainimage().handle_configure(tk.Event())

        if not maintext().is_modified():
            return True

        save = messagebox.askyesnocancel(
            title="Save document?",
            message="Save changes to document first?",
            detail="Your changes will be lost if you don't save them.",
            icon=messagebox.WARNING,
        )
        # Trap Cancel from messagebox
        if save is None:
            return False
        # Trap Cancel from save-as dialog
        if save and not self.save_file():
            return False
        return True

    def load_bin(self, basename: str) -> bool:
        """Load bin file associated with current file.

        If bin file not found, returns silently.

        Args:
            basename - name of current text file - bin file has ".bin" appended.

        Returns:
            True if no bin, or bin matches main file; False otherwise
        """
        bin_dict = load_dict_from_json(bin_name(basename))
        if bin_dict is None:
            return True
        return self.interpret_bin(bin_dict)  # type: ignore[arg-type]

    def save_bin(self, basename: str) -> None:
        """Save bin file associated with current file.

        Args:
            basename: name of current text file - bin file has ".bin" appended.
        """
        binfile_name = bin_name(basename)
        bin_dict = self.create_bin()
        with open(binfile_name, "w", encoding="utf-8") as fp:
            json.dump(bin_dict, fp, indent=2, ensure_ascii=False)

    def interpret_bin(self, bin_dict: BinDict) -> bool:
        """Interpret bin file dictionary and set necessary variables, etc.

        Args:
            bin_dict: Dictionary loaded from bin file.

        Returns:
            True if main file and text file match, False otherwise.
        """
        self.set_initial_position(bin_dict.get(BINFILE_KEY_INSERTPOS), maintext())
        self.set_initial_position(
            bin_dict.get(BINFILE_KEY_INSERTPOSPEER), maintext().peer
        )
        # Since object loaded from bin file is a dictionary of dictionaries,
        # need to create PageDetails from the loaded raw data.
        if page_details := bin_dict.get(BINFILE_KEY_PAGEDETAILS):
            for img, detail in page_details.items():
                self.page_details[img] = PageDetail(
                    detail["index"], detail["style"], detail["number"]
                )
        self.set_page_marks(self.page_details)
        self.image_dir = bin_dict.get(BINFILE_KEY_IMAGEDIR)
        self.project_id = bin_dict.get(BINFILE_KEY_PROJECTID)
        self.languages = bin_dict.get(BINFILE_KEY_LANGUAGES)
        bookmarks: Optional[dict[str, str]]
        if bookmarks := bin_dict.get(BINFILE_KEY_BOOKMARKS):
            for key, value in bookmarks.items():
                self.set_bookmark_index(key, value)

        md5checksum = bin_dict.get(BINFILE_KEY_MD5CHECKSUM)
        return not md5checksum or md5checksum == self.get_md5_checksum()

    def create_bin(self) -> BinDict:
        """From relevant variables, etc., create dictionary suitable for saving
        to bin file.

        Returns:
            Dictionary of settings to be saved in bin file
        """
        self.update_page_marks(self.page_details)
        bin_dict: BinDict = {
            BINFILE_KEY_MD5CHECKSUM: self.get_md5_checksum(),
            BINFILE_KEY_INSERTPOS: maintext().index(tk.INSERT),
            BINFILE_KEY_INSERTPOSPEER: maintext().peer.index(tk.INSERT),
            BINFILE_KEY_PAGEDETAILS: self.page_details,
            BINFILE_KEY_IMAGEDIR: self.image_dir,
            BINFILE_KEY_PROJECTID: self.project_id,
            BINFILE_KEY_LANGUAGES: self.languages,
            BINFILE_KEY_BOOKMARKS: self.get_bookmarks(),
        }
        return bin_dict

    def get_md5_checksum(self) -> str:
        """Get checksum from maintext file contents.

        Returns:
            MD5 checksum
        """
        return hashlib.md5(maintext().get_text().encode()).hexdigest()

    def store_recent_file(self, filename: str) -> None:
        """Store given filename in list of recent files.

        Args:
            filename: Name of new file to add to list.
        """
        self.remove_recent_file(filename)
        recents = preferences.get(PrefKey.RECENT_FILES)
        recents.insert(0, filename)
        del recents[NUM_RECENT_FILES:]
        preferences.set(PrefKey.RECENT_FILES, recents)
        self._filename_callback()

    def remove_recent_file(self, filename: str) -> None:
        """Remove given filename from list of recent files.

        Args:
            filename: Name of new file to add to list.
        """
        recents = preferences.get(PrefKey.RECENT_FILES)
        if filename in recents:
            recents.remove(filename)
            preferences.set(PrefKey.RECENT_FILES, recents)
        self._filename_callback()

    def set_page_marks(self, page_details: PageDetails) -> None:
        """Set page marks from keys/values in dictionary.

        Args:
            page_details: Dictionary of page details, including indexes.
        """
        self.remove_page_marks()
        if page_details:
            for img, detail in page_details.items():
                mark = page_mark_from_img(img)
                maintext().set_mark_position(mark, IndexRowCol(detail["index"]))

    def set_initial_position(self, index: str | None, widget: tk.Text) -> None:
        """Set initial cursor position after file is loaded.

        Args:
            index: Location for insert cursor. If none, go to start.
            widget: The widget to set insert cursor for.
        """
        if not index:
            index = "1.0"
        maintext().set_insert_index(IndexRowCol(index), focus_widget=widget)

    def update_page_marks(self, page_details: PageDetails) -> None:
        """Update page mark locations in page details structure.

        Args:
            page_details: Dictionary of page details, including indexes.
        """
        mark = "1.0"
        while mark := maintext().page_mark_next(mark):
            img = img_from_page_mark(mark)
            assert img in page_details
            page_details[img]["index"] = maintext().index(mark)

    def mark_page_boundaries(self) -> None:
        """Loop through whole file, ensuring all page separator lines
        are in standard format, and setting page marks at the
        start of each page separator line.
        """
        if not self.contains_page_marks():
            self.page_details = PageDetails()
        page_num_style = STYLE_ARABIC
        page_num = "1"

        pattern = re.compile(PAGE_SEPARATOR_REGEX)
        search_start = "1.0"
        while page_index := maintext().search(
            PAGE_SEPARATOR_REGEX, search_start, regexp=True, stopindex="end"
        ):
            line_start = maintext().index(page_index + " linestart")
            line_end = page_index + " lineend"
            line = maintext().get(line_start, line_end)
            # Always matches since same regex as earlier search
            if match := pattern.search(line):
                (page, ext) = match.group(1, 2)
                standard_line = f"-----File: {page}.{ext}"
                standard_line += "-" * (75 - len(standard_line))
                # Don't standarize line if it has a page marker flag on it, or you'll delete the flag!
                if line != standard_line and not re.search(PAGE_FLAG_REGEX, line):
                    maintext().delete(line_start, line_end)
                    maintext().insert(line_start, standard_line)
                page_mark = page_mark_from_img(page)
                maintext().set_mark_position(page_mark, IndexRowCol(line_start))
                try:
                    self.page_details[page]["index"] = line_start
                except KeyError:
                    self.page_details[page] = PageDetail(
                        line_start, page_num_style, page_num
                    )
                    page_num_style = STYLE_DITTO
                    page_num = NUMBER_INCREMENT

            search_start = line_end

    def remove_page_marks(self) -> None:
        """Remove any existing page marks."""
        marklist = []
        mark = "1.0"
        while mark := maintext().page_mark_next(mark):
            marklist.append(mark)
        for mark in marklist:
            maintext().mark_unset(mark)

    def contains_page_marks(self) -> bool:
        """Check whether file contains page marks.

        Returns:
            True if file contains page marks.
        """
        return maintext().page_mark_next("1.0") != ""

    def get_current_image_path(self) -> str:
        """Return the path of the image file for the page where the insert
        cursor is located.

        Returns:
            Name of the image file for the current page, or the empty string
            if unable to get image file name.
        """
        basename = maintext().get_current_image_name()
        if self.image_dir and basename:
            basename += ".png"
            path = os.path.join(self.image_dir, basename)
            if os.path.exists(path):
                return path
        return ""

    def choose_image_dir(self) -> None:
        """Allow user to select directory containing png image files"""
        self.image_dir = filedialog.askdirectory(
            mustexist=True, title=f"Select {folder_dir_str(True)} containing scans"
        )

    def get_current_page_label(self) -> str:
        """Find page label corresponding to where the insert cursor is.

        Returns:
            Page label of current page. Empty string if none found.
        """
        img = maintext().get_current_image_name()
        if img == "":
            return ""
        return self.page_details[img]["label"]

    def set_languages(self) -> None:
        """Allow the user to set languages.

        Multiple languages may be separated by any non-word character
        which will be converted to "+"
        """
        languages = simpledialog.askstring(
            "Set Language(s)", "Language(s)", parent=maintext()
        )
        if languages:
            lang_list = re.split(r"\W", languages)
            lang_list2 = [lang for lang in lang_list if lang]
            self.languages = "+".join(lang_list2)
            maintext().set_modified(True)  # Because bin file needs saving

    def goto_line(self) -> None:
        """Go to the line number the user enters"""
        line_num = simpledialog.askinteger(
            "Go To Line", "Line number", parent=maintext()
        )
        if line_num is not None:
            maintext().set_insert_index(IndexRowCol(line_num, 0))

    def goto_image(self) -> None:
        """Go to the image the user enters"""
        image_num = simpledialog.askstring(
            "Go To Page (png)", "Image number", parent=maintext()
        )
        if image_num is not None:
            image_num = re.sub(r"\.png$", "", image_num)
            self.do_goto_image(image_num)

    def do_goto_image(self, image_num: Optional[str]) -> None:
        """Go to page corresponding to the given image number.

        Args:
            image_num: Number of image to go to, e.g. "001" (can omit leading zeroes)
        """
        if image_num is None:
            return
        for n_zero in range(5):
            prefix = "0" * n_zero
            try:
                index = maintext().rowcol(page_mark_from_img(prefix + image_num))
            except tk.TclError:
                # Bad image number
                continue
            maintext().set_insert_index(index)
            return
        sound_bell()

    def goto_page(self) -> None:
        """Go to the page the user enters."""
        page_num = simpledialog.askstring(
            "Go To Page Label", "Page label", parent=maintext()
        )
        if page_num is not None:
            # If user put "Pg" before label remove it, so correctly formatted prefix can be added
            page_num = re.sub(r"^ *(Pg)? *", "", page_num)
            image_num = self.page_details.png_from_label(PAGE_LABEL_PREFIX + page_num)
            self.do_goto_image(image_num)

    def prev_page(self) -> None:
        """Go to the start of the previous page"""
        self._next_prev_page(-1)

    def next_page(self) -> None:
        """Go to the start of the next page"""
        self._next_prev_page(1)

    def _next_prev_page(self, direction: Literal[1, -1]) -> None:
        """Go to the page before/after the current one.

        Note that cursor may not move if more than one mark is at the same point.

        Args:
            direction: +1 to go to next page; -1 for previous page
        """
        insert = maintext().get_insert_index().index()
        mark = maintext().get_current_page_mark() or insert
        if mark := maintext().page_mark_next_previous(mark, direction):
            # Store mark to cope with coincident page marks
            maintext().store_page_mark(mark)
            maintext().set_insert_index(maintext().rowcol(mark))
            return
        sound_bell()

    # Note that the following code must match the equivalent code in Guiguts 1
    # or file transfer between the two will be broken.
    def add_page_flags(self) -> None:
        """Add page flag at each page boundary.

        Done in reverse order so two adjacent boundaries preserve their order.
        """
        mark: str = tk.END
        while mark := maintext().page_mark_previous(mark):
            img = img_from_page_mark(mark)
            style = self.page_details[img]["style"]
            number = self.page_details[img]["number"]
            info = PAGE_FLAG_SEP.join([PAGE_FLAG_PREFIX + img, style, number])
            maintext().insert(
                mark,
                PAGE_FLAG_START + info + PAGE_FLAG_END,
                PAGE_FLAG_TAG,
            )

    def remove_page_flags(self) -> None:
        """Remove page flags."""
        search_range = maintext().start_to_end()
        while match := maintext().find_match(
            PAGE_FLAG_REGEX,
            search_range,
            nocase=False,
            regexp=True,
            backwards=False,
        ):
            maintext().delete(
                match.rowcol.index(), match.rowcol.index() + f"+{match.count}c"
            )
            search_range = IndexRange(match.rowcol, maintext().end())

    def update_page_marks_from_flags(self) -> bool:
        """If page marker flags in file, replace all existing page marks.

        Also tag flags to highlight them.

        Returns:
            True if page marker flags were present.
        """
        mark = "1.0"
        flag_matches = maintext().find_matches(
            PAGE_FLAG_REGEX,
            maintext().start_to_end(),
            nocase=False,
            regexp=True,
        )
        if not flag_matches:
            return False

        # Remove existing page marks, if any
        self.remove_page_marks()
        self.page_details = PageDetails()
        for match in flag_matches:
            flag_text = maintext().get_match_text(match)
            extract = re.fullmatch(PAGE_FLAG_REGEX, flag_text)
            # Will always match because same regex as above
            assert extract is not None
            img = extract[1]
            style = extract[2]
            number = extract[3]
            mark = page_mark_from_img(img)
            maintext().set_mark_position(mark, match.rowcol)
            self.page_details[img] = PageDetail(match.rowcol.index(), style, number)
            maintext().tag_add(
                PAGE_FLAG_TAG,
                match.rowcol.index(),
                match.rowcol.index() + f"+{match.count}c",
            )
        maintext().set_modified(True)
        return True

    def update_page_marks_from_ppgen(self) -> bool:
        """If `.bn` ppgen commands in file, replace all existing page marks.

        Command is of the form: `.bn a001.png // optional comment`
        We only want `a001`

        Returns:
            True if `.bn` commands were present.
        """
        regex = r"^\.bn +(.+?)\.png"
        flag_matches = maintext().find_matches(
            regex,
            maintext().start_to_end(),
            nocase=False,
            regexp=True,
        )
        if not flag_matches:
            return False

        # Remove existing page marks, if any
        self.remove_page_marks()
        self.page_details = PageDetails()
        # First page is Arabic, page 1
        style = STYLE_ARABIC
        number = "1"
        for match in flag_matches:
            bn_text = maintext().get_match_text(match)
            extract = re.fullmatch(regex, bn_text)
            # Will always match because same regex as above
            assert extract is not None
            img = extract[1]
            # Set mark and store page info
            maintext().set_mark_position(page_mark_from_img(img), match.rowcol)
            self.page_details[img] = PageDetail(match.rowcol.index(), style, number)
            # Subsequent pages are same style as first, page numbers increment
            style = STYLE_DITTO
            number = NUMBER_INCREMENT
        maintext().set_modified(True)
        return True

    def add_good_and_bad_words(self) -> None:
        """Load the words from the good and bad words files into the project dictionary."""

        gw_load = self.project_dict.add_good_bad_words(
            self.filename, load_good_words=True
        )
        bw_load = self.project_dict.add_good_bad_words(
            self.filename, load_good_words=False
        )
        if gw_load or bw_load:
            self.project_dict.save(self.filename)
        else:
            logger.error(
                f"Neither {GOOD_WORDS_FILENAME} nor {BAD_WORDS_FILENAME} was found"
            )

    def add_good_word_to_project_dictionary(self, word: str) -> None:
        """Add a good word to the project dictionary. Save if word was added.

        Args:
            word: The word to be added.
        """
        if self.project_dict.add_good_word(word):
            self.project_dict.save(self.filename)

    def add_good_word_to_global_dictionary(self, word: str) -> None:
        """Add a good word to the user global dictionary for the current language.

        Args:
            word: The word to be added.
        """
        main_lang = maintext().get_language_list()[0]
        path = Path(preferences.prefsdir, f"dict_{main_lang}_user.txt")
        with path.open("a", encoding="utf-8") as fp:
            fp.write(f"{word}\n")

    def rewrap_selection(self) -> None:
        """Wrap selected text."""
        ranges = maintext().selected_ranges()
        if not ranges:
            sound_bell()
            return
        # Adjust selection to begin & end at start of lines
        ranges[0].start.col = min(ranges[0].start.col, 0)
        if ranges[0].end.col > 0:
            ranges[0].end.col = 0
            ranges[0].end.row += 1
        self.rewrap_section(ranges[0])

    def rewrap_all(self) -> None:
        """Wrap whole text."""
        self.rewrap_section(IndexRange("1.0", maintext().index(tk.END)))

    def rewrap_section(self, section_range: IndexRange) -> None:
        """Wrap a section of the text, preserving page mark locations.

        Args:
            section_range: Range of text to be wrapped.
        """
        maintext().selection_ranges_store_with_marks()
        maintext().undo_block_begin()

        # Dummy insert & delete, so that if user undoes the wrap, the insert
        # cursor returns to its previous point, not the first pin page mark position.
        insert_index = maintext().get_insert_index()
        maintext().insert(insert_index.index(), " ")
        maintext().delete(insert_index.index())

        maintext().strip_end_of_line_spaces()
        mark_list = self.pin_page_marks()
        maintext().rewrap_section(
            section_range, lambda: self.unpin_page_marks(mark_list)
        )
        maintext().selection_ranges_restore_from_marks()

    def rewrap_cleanup(self) -> None:
        """Clean up all rewrap markers."""
        match_regex = r"^/[#$*FILPXCR]|^[#$*FILPXCR]/"
        search_range = maintext().start_to_end()
        maintext().undo_block_begin()
        while beg_match := maintext().find_match(
            match_regex, search_range, regexp=True, nocase=True
        ):
            # If a start rewrap is followed immediately by a blank line
            # then another start rewrap, delete the blank line as well.
            beg_idx = beg_match.rowcol.index()
            is_start = maintext().get(beg_idx) == "/"
            test_txt = maintext().get(f"{beg_idx} +1l", f"{beg_idx} +2l lineend")
            nlines = 2 if is_start and re.match(r"\n/[#$*FILPXCR]", test_txt) else 1

            maintext().delete(beg_idx, f"{beg_idx} +{nlines}l")
            search_range = IndexRange(beg_match.rowcol, maintext().end())

    def pin_page_marks(self) -> list[str]:
        """Pin the page marks to locations in the text, by inserting a special
        character at each page mark.

        This allows operations such as rewrapping to happen without losing
        page mark positions. Marks are unpinned by `unpin_page_marks`.

        Returns:
            List of page mark names in reverse order to pass to `unpin_page_marks`.
        """
        # Ensure no pins left around from previous wraps - should never happen,
        # but if it has, we want to clear them.
        found: str = tk.END
        while found := maintext().search(PAGEMARK_PIN, found, backwards=True):
            maintext().delete(found, f"{found} +1c")

        mark_list = []
        mark: str = tk.END
        while mark := maintext().page_mark_previous(mark):
            mark_list.append(mark)
            maintext().insert(mark, PAGEMARK_PIN)
        return mark_list

    def unpin_page_marks(self, mark_list: list[str]) -> None:
        """Unpin the page marks by setting them to the locations of the special
        characters inserted by `pin_page_marks`.

        Args:
            mark_list: List of page mark names in reverse order from `pin_page_marks`.
        """
        # First set each page mark to the pinned location - this avoids setting two
        # page marks adjacent to one another (due to the pin character) because their
        # order isn't guaranteed if you do that, so they can end up swapped.
        found = "1.0"
        while found := maintext().search(PAGEMARK_PIN, found, "end"):
            maintext().mark_set(mark_list.pop(), found)
            found = f"{found} +1c"
        # Now it's safe to remove the pins, in reverse order to simplify loop
        found = tk.END
        while found := maintext().search(PAGEMARK_PIN, found, backwards=True):
            maintext().delete(found, f"{found} +1c")

    def set_bookmark_index(self, bm_name: str, idx: str) -> None:
        """Set the position of a bookmark via name & index string.

        Args:
            bm_name: Name of bookmark to set (assumed valid).
            idx: Index of bookmark position in file.
        """
        maintext().set_mark_position(bm_name, maintext().rowcol(idx))

    def set_bookmark(self, bm_num: int) -> None:
        """Set the position of the given bookmark at the current insert location.

        Args:
            bm_num: Number of bookmark to add (must be 1 to 5).
        """
        assert 1 <= bm_num <= 5
        maintext().set_mark_position(
            f"{BOOKMARK_BASE}{bm_num}", maintext().get_insert_index()
        )
        self.highlight_bookmark(bm_num)

    def goto_bookmark(self, bm_num: int) -> None:
        """Set the insert position to the location of the given bookmark.

        Args:
            bm_num: Number of bookmark to find (must be 1 to 5).
        """
        assert 1 <= bm_num <= 5
        try:
            maintext().set_insert_index(
                maintext().rowcol(f"{BOOKMARK_BASE}{bm_num}"), focus=False
            )
        except tk.TclError:
            sound_bell()  # Bookmark hasn't been set
            return
        self.highlight_bookmark(bm_num)

    def highlight_bookmark(self, bm_num: int) -> None:
        """Temporarily highlight bookmark position.

        Args:
            bm_num: Number of bookmark to find (must be 1 to 5).
        """
        assert 1 <= bm_num <= 5
        maintext().tag_add(BOOKMARK_TAG, maintext().index(f"{BOOKMARK_BASE}{bm_num}"))
        maintext().after(1000, self.remove_bookmark_tags)

    def get_bookmarks(self) -> dict[str, str]:
        """Get dictionary of bookmark names:locations.

        Assumes if a mark begins with `BOOKMARK_BASE` it's a bookmark.
        """
        bookmarks: dict[str, str] = {}
        for name in maintext().mark_names():
            if name.startswith(BOOKMARK_BASE):
                assert re.fullmatch(f"{BOOKMARK_BASE}[1-5]", name)
                bookmarks[name] = maintext().index(name)
        return bookmarks

    def remove_bookmark_tags(self) -> None:
        """Remove all bookmark highlightling."""
        maintext().tag_remove(BOOKMARK_TAG, "1.0", tk.END)


def bin_name(basename: str) -> str:
    """Get the name of the bin file associated with a text file.

    Args:
        basename: Name of text file.

    Returns:
        Name of associated bin file.
    """
    return basename + BINFILE_SUFFIX


# For convenient access, store the single File instance here,
# with a function to set/query it.
_the_file = None  # pylint: disable=invalid-name


def the_file(file: Optional[File] = None) -> File:
    """Store and return the single File widget"""
    global _the_file
    if file is not None:
        assert _the_file is None
        _the_file = file
    assert _the_file is not None
    return _the_file



"""Search/Replace functionality"""

import logging
import tkinter as tk
from tkinter import ttk
from tkinter import font as tk_font
import traceback
from typing import Any, Tuple, Optional

import regex as re

from guiguts.checkers import CheckerDialog
from guiguts.maintext import maintext, TclRegexCompileError, FindMatch
from guiguts.preferences import preferences, PersistentBoolean, PrefKey
from guiguts.utilities import sound_bell, IndexRowCol, IndexRange, sing_plur
from guiguts.widgets import (
    ToplevelDialog,
    Combobox,
    mouse_bind,
    register_focus_widget,
    process_accel,
    Busy,
)

logger = logging.getLogger(__package__)

MARK_FOUND_START = "FoundStart"
MARK_FOUND_END = "FoundEnd"
MARK_END_RANGE = "SearchRangeEnd"
PADX = 2
PADY = 2


class NoMatchFoundError(Exception):
    """Raised when no match is found for the search string."""


class SearchDialog(ToplevelDialog):
    """A Toplevel dialog that allows the user to search/replace.

    Attributes:
        reverse: True to search backwards.
        matchcase: True to ignore case.
        wrap: True to wrap search round beginning/end of file.
        regex: True to use regex search.
        selection: True to restrict counting, replacing, etc., to selected text.
    """

    manual_page = "Searching"
    # Cannot be initialized here, since Tk root may not be created yet
    selection: tk.BooleanVar

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize Search dialog."""

        # Initialize "Selection" variable on first instantiation.
        # Persistent only during this run of the program.
        try:
            SearchDialog.selection
        except AttributeError:
            SearchDialog.selection = tk.BooleanVar(value=False)
        kwargs["resize_y"] = False
        kwargs["disable_geometry_save"] = True
        super().__init__("Search & Replace", *args, **kwargs)
        self.minsize(400, 100)

        # Frames
        self.top_frame.columnconfigure(0, weight=1)
        for row in range(5):
            self.top_frame.rowconfigure(row, weight=0)
        self.top_frame.rowconfigure(6, weight=1)
        options_frame = ttk.Frame(
            self.top_frame, padding=3, borderwidth=1, relief=tk.GROOVE
        )
        options_frame.grid(
            row=0, column=0, columnspan=3, rowspan=2, ipady=5, sticky="NSEW"
        )
        options_frame.columnconfigure(0, weight=1)
        options_frame.columnconfigure(1, weight=1)
        options_frame.columnconfigure(2, weight=1)
        options_frame.rowconfigure(0, weight=1)
        options_frame.rowconfigure(1, weight=1)
        message_frame = ttk.Frame(self.top_frame, padding=1)
        message_frame.grid(row=6, column=0, columnspan=5, sticky="NSEW", pady=(5, 0))
        self.separator = ttk.Separator(self.top_frame, orient=tk.VERTICAL)
        self.separator.grid(row=0, column=3, rowspan=6, padx=2, sticky="NSEW")

        # Search
        style = ttk.Style()
        new_col = "#ff8080" if maintext().is_dark_theme() else "#e60000"
        style.configure("BadRegex.TCombobox", foreground=new_col)

        def is_valid_regex(new_value: str) -> bool:
            """Validation routine for Search Combobox - check value is a valid regex.

            Note that it always returns True because we want user to be able to type
            the character. It just alerts the user by switching to the BadRegex style.
            """
            if preferences.get(PrefKey.SEARCHDIALOG_REGEX):
                try:
                    re.compile(new_value)
                    self.search_box["style"] = ""
                except re.error:
                    self.search_box["style"] = "BadRegex.TCombobox"
            else:
                self.search_box["style"] = ""
            return True

        self.font = tk_font.Font(
            family=maintext().font.cget("family"),
            size=maintext().font.cget("size"),
        )
        self.search_box = Combobox(
            self.top_frame,
            PrefKey.SEARCH_HISTORY,
            width=30,
            font=self.font,
            validate=tk.ALL,
            validatecommand=(self.register(is_valid_regex), "%P"),
        )
        self.search_box.grid(row=2, column=0, padx=PADX, pady=PADY, sticky="NSEW")
        # Register search box to have its focus tracked for inserting special characters
        register_focus_widget(self.search_box)
        self.search_box.focus()

        search_button = ttk.Button(
            self.top_frame,
            text="Search",
            default="active",
            takefocus=False,
            command=self.search_clicked,
        )
        search_button.grid(row=2, column=1, padx=PADX, pady=PADY, sticky="NSEW")
        mouse_bind(
            search_button,
            "Shift+1",
            lambda *args: self.search_clicked(opposite_dir=True),
        )
        self.bind("<Return>", lambda *args: self.search_clicked())
        self.bind(
            "<Shift-Return>", lambda *args: self.search_clicked(opposite_dir=True)
        )

        # First/Last button - find first/last occurrence in file
        first_button = ttk.Button(
            self.top_frame,
            text="Last" if preferences.get(PrefKey.SEARCHDIALOG_REVERSE) else "First",
            takefocus=False,
            command=lambda *args: self.search_clicked(first_last=True),
        )
        first_button.grid(row=2, column=2, padx=PADX, pady=PADY, sticky="NSEW")

        # Count & Find All
        self.count_btn = ttk.Button(
            self.top_frame,
            text="Count",
            takefocus=False,
            command=self.count_clicked,
        )
        self.count_btn.grid(row=1, column=4, padx=PADX, pady=PADY, sticky="NSEW")
        ttk.Button(
            self.top_frame,
            text="Find All",
            takefocus=False,
            command=self.findall_clicked,
        ).grid(row=2, column=4, padx=PADX, pady=PADY, sticky="NSEW")

        def set_first_last() -> None:
            """Set text in First/Last button depending on direction."""
            first_button["text"] = (
                "Last" if preferences.get(PrefKey.SEARCHDIALOG_REVERSE) else "First",
            )

        # Options
        ttk.Checkbutton(
            options_frame,
            text="Reverse",
            variable=PersistentBoolean(PrefKey.SEARCHDIALOG_REVERSE),
            command=set_first_last,
            takefocus=False,
        ).grid(row=0, column=0, padx=2, sticky="NSEW")
        ttk.Checkbutton(
            options_frame,
            text="Match case",
            variable=PersistentBoolean(PrefKey.SEARCHDIALOG_MATCH_CASE),
            takefocus=False,
        ).grid(row=0, column=1, padx=2, sticky="NSEW")
        ttk.Checkbutton(
            options_frame,
            text="Regex",
            variable=PersistentBoolean(PrefKey.SEARCHDIALOG_REGEX),
            takefocus=False,
            command=lambda: is_valid_regex(self.search_box.get()),
        ).grid(row=0, column=2, padx=2, sticky="NSEW")
        ttk.Checkbutton(
            options_frame,
            text="Whole word",
            variable=PersistentBoolean(PrefKey.SEARCHDIALOG_WHOLE_WORD),
            takefocus=False,
        ).grid(row=1, column=0, padx=2, sticky="NSEW")
        ttk.Checkbutton(
            options_frame,
            text="Wrap around",
            variable=PersistentBoolean(PrefKey.SEARCHDIALOG_WRAP),
            takefocus=False,
        ).grid(row=1, column=1, padx=2, sticky="NSEW")
        ttk.Checkbutton(
            options_frame,
            text="Multi-replace",
            variable=PersistentBoolean(PrefKey.SEARCHDIALOG_MULTI_REPLACE),
            takefocus=False,
            command=lambda: self.show_multi_replace(
                preferences.get(PrefKey.SEARCHDIALOG_MULTI_REPLACE)
            ),
        ).grid(row=1, column=2, padx=2, sticky="NSEW")
        ttk.Checkbutton(
            self.top_frame,
            text="In selection",
            variable=SearchDialog.selection,
            takefocus=False,
        ).grid(row=0, column=4, sticky="NSE")

        # Replace
        self.replace_box: list[Combobox] = []
        self.replace_btn: list[ttk.Button] = []
        self.rands_btn: list[ttk.Button] = []
        self.repl_all_btn: list[ttk.Button] = []
        for rep_num in range(3):
            cbox = Combobox(
                self.top_frame, PrefKey.REPLACE_HISTORY, width=30, font=self.font
            )
            cbox.grid(row=rep_num + 3, column=0, padx=PADX, pady=PADY, sticky="NSEW")
            self.replace_box.append(cbox)
            # Register replace box to have its focus tracked for inserting special characters
            register_focus_widget(cbox)

            r_btn = ttk.Button(
                self.top_frame,
                text="Replace",
                takefocus=False,
                command=lambda idx=rep_num: self.replace_clicked(idx),  # type: ignore[misc]
            )
            r_btn.grid(row=rep_num + 3, column=1, padx=PADX, pady=PADY, sticky="NSEW")
            self.replace_btn.append(r_btn)

            rands_button = ttk.Button(
                self.top_frame,
                text="R & S",
                takefocus=False,
                command=lambda idx=rep_num: self.replace_clicked(  # type: ignore[misc]
                    idx, search_again=True
                ),
            )
            rands_button.grid(
                row=rep_num + 3, column=2, padx=PADX, pady=PADY, sticky="NSEW"
            )
            mouse_bind(
                rands_button,
                "Shift+1",
                lambda _e, idx=rep_num: self.replace_clicked(  # type: ignore[misc]
                    idx, opposite_dir=True, search_again=True
                ),
            )
            self.rands_btn.append(rands_button)

            repl_all_btn = ttk.Button(
                self.top_frame,
                text="Replace All",
                takefocus=False,
                command=lambda idx=rep_num: self.replaceall_clicked(idx),  # type: ignore[misc]
            )
            repl_all_btn.grid(
                row=rep_num + 3, column=4, padx=PADX, pady=PADY, sticky="NSEW"
            )
            self.repl_all_btn.append(repl_all_btn)
        _, key_event = process_accel("Cmd/Ctrl+Return")
        self.bind(key_event, lambda *args: self.replace_clicked(0, search_again=True))
        _, key_event = process_accel("Cmd/Ctrl+Shift+Return")
        self.bind(
            key_event,
            lambda *args: self.replace_clicked(0, opposite_dir=True, search_again=True),
        )

        # Message (e.g. count)
        message_frame.columnconfigure(0, weight=1)
        self.message = ttk.Label(
            message_frame, borderwidth=1, relief="sunken", padding=5
        )
        self.message.grid(row=0, column=0, sticky="NSEW")

        self.show_multi_replace(
            preferences.get(PrefKey.SEARCHDIALOG_MULTI_REPLACE), resize=False
        )

        # Now dialog geometry is set up, set width to user pref, leaving height as it is
        self.config_width()
        self.allow_geometry_save()

    def reset(self) -> None:
        """Called when dialog is reset/destroyed - remove search highlights."""
        maintext().highlight_search_deactivate()

    def show_multi_replace(self, show: bool, resize: bool = True) -> None:
        """Show or hide the multi-replace buttons.

        Args:
            show: True to show the extra replace buttons.
            resize: True (default) to grow/shrink dialog to take account of show/hide
                When dialog first created, its size is stored in prefs, so won't need resize
        """
        for w_list in (
            self.replace_box,
            self.replace_btn,
            self.rands_btn,
            self.repl_all_btn,
        ):
            for widget in w_list[1:]:
                if show:
                    widget.grid()
                else:
                    widget.grid_remove()
        self.separator.grid(rowspan=6 if show else 4)

        if not resize:
            return

        # Height needs to grow/shrink by the space taken up by 2 entry fields
        offset = 2 * (self.replace_box[0].winfo_y() - self.search_box.winfo_y())
        geometry = self.geometry()
        height = int(re.sub(r"\d+x(\d+).+", r"\1", geometry))
        height += offset if show else -offset
        geometry = re.sub(r"(\d+x)\d+(.+)", rf"\g<1>{height}\g<2>", geometry)
        self.geometry(geometry)

    def search_box_set(self, search_string: str) -> None:
        """Set string in search box.

        Also selects the string, and places the cursor at the end

        Args:
            search_string: String to put in search box.
        """
        self.search_box.set(search_string)
        self.search_box.select_range(0, tk.END)
        self.search_box.icursor(tk.END)
        self.search_box.focus()

    def search_clicked(
        self, opposite_dir: bool = False, first_last: bool = False
    ) -> str:
        """Search for the string in the search box.

        Args:
            opposite_dir: True to search in opposite direction to reverse flag setting
            first_last: True to begin search at start/end of file
        Returns:
            "break" to avoid calling other callbacks
        """
        search_string = self.search_box.get()
        if not search_string:
            return "break"
        self.search_box.add_to_history(search_string)

        # "Reverse flag XOR Shift-key" searches backwards
        backwards = preferences.get(PrefKey.SEARCHDIALOG_REVERSE) ^ opposite_dir
        if first_last:
            start_rowcol = (
                maintext().end()
                if preferences.get(PrefKey.SEARCHDIALOG_REVERSE)
                else maintext().start()
            )
        else:
            start_rowcol = get_search_start(backwards)
        stop_rowcol = maintext().start() if backwards else maintext().end()
        message = ""

        try:
            maintext().search_pattern = self.search_box.get()
            maintext().search_highlight_active.set(True)

            # Now that "background" matches are highlighted, find the next match
            # and jump there as the "active" match. Uses the "sel" highlight.
            _do_find_next(
                search_string, backwards, IndexRange(start_rowcol, stop_rowcol)
            )
        except re.error as e:
            message = message_from_regex_exception(e)
        except NoMatchFoundError:
            message = "No matches found"
        self.display_message(message)
        return "break"

    def search_forwards(self) -> str:
        """Force forward search regardless of reverse flag.

        Returns:
            "break" to avoid calling other callbacks
        """
        self.search_clicked(opposite_dir=preferences.get(PrefKey.SEARCHDIALOG_REVERSE))
        return "break"

    def search_backwards(self) -> str:
        """Force backward search regardless of reverse flag.

        Returns:
            "break" to avoid calling other callbacks
        """
        self.search_clicked(
            opposite_dir=not preferences.get(PrefKey.SEARCHDIALOG_REVERSE)
        )
        return "break"

    def count_clicked(self) -> Optional[list[FindMatch]]:
        """Count how many times search string occurs in file (or selection).

        Display count in Search dialog.

        Returns:
            List of FindMatch objects (None if error).
        """
        search_string = self.search_box.get()
        if not search_string:
            return None
        self.search_box.add_to_history(search_string)

        find_range, range_name = get_search_range()
        if find_range is None:
            self.display_message('No text selected for "In selection" find')
            sound_bell()
            return None

        try:
            matches = maintext().find_all(find_range, search_string)
        except re.error as e:
            self.display_message(message_from_regex_exception(e))
            return None
        count = len(matches)
        match_str = sing_plur(count, "match", "matches")
        self.display_message(f"Found: {match_str} {range_name}")
        return matches

    def findall_clicked(self) -> None:
        """Callback when Find All button clicked.

        Find & count occurrences, then display in dialog.
        """
        matches = self.count_clicked()
        if matches is None:
            return

        class FindAllCheckerDialog(CheckerDialog):
            """Find All dialog."""

            manual_page = "Searching#Find_All"

            def __init__(self, **kwargs: Any) -> None:
                """Initialize Find All dialog."""

                super().__init__(
                    "Search Results",
                    tooltip="\n".join(
                        [
                            "Left click: Select & find string",
                            "Right click: Remove string from this list",
                            "Shift Right click: Remove all occurrences of string from this list",
                        ]
                    ),
                    **kwargs,
                )

        checker_dialog = FindAllCheckerDialog.show_dialog(
            rerun_command=self.findall_clicked,
        )
        if not checker_dialog.winfo_exists() or not self.winfo_exists():
            Busy.unbusy()
            return

        # Construct opening line describing the search
        desc_reg = "regex" if preferences.get(PrefKey.SEARCHDIALOG_REGEX) else "string"
        prefix = f'Search for {desc_reg} "'
        desc = f'{prefix}{self.search_box.get()}"'
        if preferences.get(PrefKey.SEARCHDIALOG_MATCH_CASE):
            desc += ", matching case"
        if preferences.get(PrefKey.SEARCHDIALOG_WHOLE_WORD):
            desc += ", whole words only"
        if SearchDialog.selection.get():
            desc += ", within selection"
        checker_dialog.add_header(desc, "")

        for match in matches:
            line = maintext().get(
                f"{match.rowcol.index()} linestart",
                f"{match.rowcol.index()}+{match.count}c lineend",
            )
            end_rowcol = IndexRowCol(
                maintext().index(match.rowcol.index() + f"+{match.count}c")
            )
            hilite_start = match.rowcol.col
            # If multiline, lines will be concatenated, so adjust end hilite point
            if end_rowcol.row > match.rowcol.row:
                not_matched = maintext().get(
                    f"{match.rowcol.index()}+{match.count}c",
                    f"{match.rowcol.index()}+{match.count}c lineend",
                )
                hilite_end = len(line) - len(not_matched)
            else:
                hilite_end = end_rowcol.col
            checker_dialog.add_entry(
                line, IndexRange(match.rowcol, end_rowcol), hilite_start, hilite_end
            )
        checker_dialog.add_footer("", "End of search results")
        checker_dialog.display_entries()

    def replace_clicked(
        self, box_num: int, opposite_dir: bool = False, search_again: bool = False
    ) -> str:
        """Replace the found string with the replacement in the replace box.

        Args:
            box_num: Which replace box's Replace button was clicked.
            opposite_dir: True to go in opposite direction to the "Reverse" flag.
            search_again: True to find next match after replacement.

        Returns:
            "break" to avoid calling other callbacks
        """
        search_string = self.search_box.get()
        self.search_box.add_to_history(search_string)
        replace_string = self.replace_box[box_num].get()
        self.replace_box[box_num].add_to_history(replace_string)

        try:
            start_index = maintext().index(MARK_FOUND_START)
            end_index = maintext().index(MARK_FOUND_END)
        except tk.TclError:
            # If Replace & Search, then even if we can't Replace, do a Search
            if search_again:
                self.search_clicked(opposite_dir=opposite_dir)
            else:
                sound_bell()
                self.display_message("No text found to replace")
            return "break"

        match_text = maintext().get(start_index, end_index)
        if preferences.get(PrefKey.SEARCHDIALOG_REGEX):
            flags = (
                0 if preferences.get(PrefKey.SEARCHDIALOG_MATCH_CASE) else re.IGNORECASE
            )
            try:
                replace_string = get_regex_replacement(
                    search_string, replace_string, match_text, flags=flags
                )
            except re.error as e:
                self.display_message(f"Regex error: {str(e)}")
                sound_bell()
                return "break"
        maintext().undo_block_begin()
        maintext().replace(start_index, end_index, replace_string)
        # "Reverse flag XOR Shift-key" searches backwards
        backwards = preferences.get(PrefKey.SEARCHDIALOG_REVERSE) ^ opposite_dir
        # Ensure cursor is at correct end of replaced string - depends on direction.
        maintext().set_insert_index(
            maintext().rowcol(MARK_FOUND_START if backwards else MARK_FOUND_END),
            focus=False,
        )
        maintext().mark_unset(MARK_FOUND_START, MARK_FOUND_END)
        maintext().clear_selection()
        if search_again:
            find_next(backwards=backwards)
        return "break"

    def replaceall_clicked(self, box_num: int) -> None:
        """Callback when Replace All button clicked.

        Replace in whole file or just in selection.

        Args:
            box_num: Which replace box's Replace button was clicked.

        """
        search_string = self.search_box.get()
        if not search_string:
            return
        self.search_box.add_to_history(search_string)
        replace_string = self.replace_box[box_num].get()
        self.replace_box[box_num].add_to_history(replace_string)

        replace_range, range_name = get_search_range()
        if replace_range is None:
            self.display_message('No text selected for "In selection" replace')
            sound_bell()
            return

        if SearchDialog.selection.get():
            maintext().selection_ranges_store_with_marks()
        maintext().undo_block_begin()

        regexp = preferences.get(PrefKey.SEARCHDIALOG_REGEX)
        replace_match = replace_string

        try:
            matches = maintext().find_all(replace_range, search_string)
        except re.error as e:
            self.display_message(message_from_regex_exception(e))
            return

        flags = 0 if preferences.get(PrefKey.SEARCHDIALOG_MATCH_CASE) else re.IGNORECASE

        # Work backwards so replacements don't affect future match locations
        for match in reversed(matches):
            start_index = match.rowcol.index()
            end_index = maintext().index(start_index + f"+{match.count}c")
            match_text = maintext().get(start_index, end_index)
            if regexp:
                try:
                    replace_match = get_regex_replacement(
                        search_string, replace_string, match_text, flags=flags
                    )
                except re.error as e:
                    self.display_message(f"Regex error: {str(e)}")
                    sound_bell()
                    return
            maintext().replace(start_index, end_index, replace_match)

        if SearchDialog.selection.get():
            maintext().selection_ranges_restore_from_marks()
        else:
            maintext().clear_selection()

        match_str = sing_plur(len(matches), "match", "matches")
        self.display_message(f"Replaced: {match_str} {range_name}")

    def display_message(self, message: str = "") -> None:
        """Display message in Search dialog.

        Args:
            message: Message to be displayed - clears message if arg omitted
        """
        if self.message.winfo_exists():
            self.message["text"] = message


def show_search_dialog() -> None:
    """Show the Search dialog and set the string in search box
    to the selected text if any (up to first newline)."""
    dlg = SearchDialog.show_dialog()
    dlg.search_box_set(maintext().selected_text().split("\n", 1)[0])
    dlg.display_message()


def find_next(backwards: bool = False) -> None:
    """Find next occurrence of most recent search string.

    Takes account of current wrap, matchcase, regex & wholeword flag settings
    in Search dialog. If dialog hasn't been shown previously or there is
    no recent search string sounds bell and returns.

    Args:
        backwards: True to search backwards (not dependent on "Reverse"
            setting in dialog).
    """
    try:
        SearchDialog.selection
    except AttributeError:
        sound_bell()
        return  # Dialog has never been instantiated

    search_string = ""
    # If dialog is visible, then string in search box takes priority over
    # previously-searched-for string.
    if dlg := SearchDialog.get_dialog():
        search_string = dlg.search_box.get()
        dlg.search_box.add_to_history(search_string)
        dlg.display_message("")
    if not search_string:
        try:
            search_string = preferences.get(PrefKey.SEARCH_HISTORY)[0]
        except IndexError:
            sound_bell()
            return  # No Search History

    start_rowcol = get_search_start(backwards)
    stop_rowcol = maintext().start() if backwards else maintext().end()
    try:
        _do_find_next(search_string, backwards, IndexRange(start_rowcol, stop_rowcol))
    except TclRegexCompileError as exc:
        logger.error(str(exc))
    except NoMatchFoundError:
        if dlg:
            dlg.display_message("No more matches found")


def _do_find_next(
    search_string: str, backwards: bool, search_limits: IndexRange
) -> None:
    """Find next occurrence of string from start_point.

    Args:
        search_string: String to search for.
        backwards: True to search backwards.
        start_point: Point to search from.
    """
    match = maintext().find_match_user(
        search_string,
        search_limits.start,
        nocase=not preferences.get(PrefKey.SEARCHDIALOG_MATCH_CASE),
        wholeword=preferences.get(PrefKey.SEARCHDIALOG_WHOLE_WORD),
        regexp=preferences.get(PrefKey.SEARCHDIALOG_REGEX),
        backwards=backwards,
        wrap=preferences.get(PrefKey.SEARCHDIALOG_WRAP),
    )

    if match:
        rowcol_end = maintext().rowcol(match.rowcol.index() + f"+{match.count}c")
        maintext().set_insert_index(match.rowcol, focus=False)
        maintext().do_select(IndexRange(match.rowcol, rowcol_end))
        maintext().set_mark_position(MARK_FOUND_START, match.rowcol, gravity=tk.LEFT)
        maintext().set_mark_position(MARK_FOUND_END, rowcol_end, gravity=tk.RIGHT)
    else:
        maintext().highlight_search_deactivate()
        sound_bell()
        raise NoMatchFoundError


def get_search_start(backwards: bool) -> IndexRowCol:
    """Find point to start searching from.

    Start from current insert point unless following are true:
    We are searching forward;
    Current insert point is at start of previously found match;
    Start of previous match is still selected (or it was a zero-length match)
    If all are true, advance to end of match.

    Additionally, searching for zero-length matches when already at start
    or end of file, needs special handling

    Args:
        backwards: True if searching backwards.
    """
    start_rowcol = maintext().get_insert_index()
    start_index = start_rowcol.index()
    try:
        at_previous_match = maintext().compare(MARK_FOUND_START, "==", start_index)
    except tk.TclError:
        at_previous_match = False  # MARK not found
    # We've previously done a search, and are now doing another, so various special
    # cases needed to avoid getting stuck at a match
    if at_previous_match:
        zero_len = maintext().compare(MARK_FOUND_START, "==", MARK_FOUND_END)
        if backwards:
            if zero_len:
                # If at start of file, and wrapping, then next reverse search is from end,
                # otherwise, just go back one character
                if preferences.get(PrefKey.SEARCHDIALOG_WRAP) and maintext().compare(
                    MARK_FOUND_START, "==", "1.0"
                ):
                    start_rowcol = maintext().end()
                else:
                    start_rowcol = maintext().rowcol(start_index + "-1c")
        else:
            if zero_len:
                # If at end of file, and wrapping, then next search is from start,
                # otherwise, just go forward one character
                if preferences.get(PrefKey.SEARCHDIALOG_WRAP) and maintext().compare(
                    MARK_FOUND_START, "==", maintext().end().index()
                ):
                    start_rowcol = maintext().start()
                else:
                    start_rowcol = maintext().rowcol(start_index + "+1c")
            elif sel_ranges := maintext().selected_ranges():
                if maintext().compare(sel_ranges[0].start.index(), "==", start_index):
                    start_rowcol = maintext().rowcol(MARK_FOUND_END)
    return start_rowcol


def get_regex_replacement(
    search_regex: str,
    replace_regex: str,
    match_text: str,
    flags: int,
) -> str:
    """Find actual replacement string, given the search & replace regexes
    and the matching text.

    Raises re.error exception if regexes are bad

    Args:
        search_regex: Regex that was used for search
        replace_regex: Regex used for replacement
        match_text: The text that was actually matched
        flags: "re.sub" flags to pass when performing the regex substitution

    Returns:
        Replacement string.
    """
    temp_bs = "\x9f"
    # Unused character to temporarily replace backslash in `\C`, `\E`

    # Since below we do a sub on the match text, rather than the whole text, we need
    # to handle start/end word boundaries and look-behind/ahead by removing them.
    # At some point the sub will be done manually, handing groups, execution of
    # python code, etc., like in GG1. At that point, these fixes can probably go.
    search_regex = re.sub(r"^\(\?<=.*?\)", "", search_regex)
    search_regex = re.sub(r"\(\?=.*?\)$", "", search_regex)
    search_regex = search_regex.removeprefix(r"\b").removesuffix(r"\b")

    # Allow execution of python code
    replace_regex = replace_regex.replace(r"\C", f"{temp_bs}C")
    replace_regex = replace_regex.replace(r"\E", f"{temp_bs}E")
    replace_str = re.sub(search_regex, replace_regex, match_text, flags=flags)
    if f"{temp_bs}C" in replace_str:
        while True:
            c_index = replace_str.find(f"{temp_bs}C")
            if c_index < 0:
                break
            e_index = replace_str.find(f"{temp_bs}E", c_index)
            if e_index < 0:
                break
            python_in = replace_str[c_index + 2 : e_index]
            try:
                python_out = str(eval(python_in))  # pylint:disable=eval-used
            except Exception as exc:
                tb = re.sub(
                    r'.+File "<string>", line 1[^\n]*',
                    r"\\C...\\E - error in Python code",
                    traceback.format_exc(),
                    flags=re.DOTALL,
                )
                logger.error(tb)
                raise re.error("\\C...\\E error - see message log for details") from exc
            replace_str = (
                replace_str[:c_index] + python_out + replace_str[e_index + 2 :]
            )
    return replace_str


def get_search_range() -> Tuple[Optional[IndexRange], str]:
    """Get range to search over, based on checkbox settings.

    Returns:
        Range to search over, and string to describe the range.
    """
    replace_range = None
    range_name = ""
    if SearchDialog.selection.get():
        range_name = "in selection"
        if sel_ranges := maintext().selected_ranges():
            replace_range = sel_ranges[0]
    else:
        if preferences.get(PrefKey.SEARCHDIALOG_WRAP):
            range_name = "in entire file"
            replace_range = maintext().start_to_end()
        elif preferences.get(PrefKey.SEARCHDIALOG_REVERSE):
            range_name = "from start of file to current location"
            replace_range = IndexRange(
                maintext().start(), maintext().get_insert_index()
            )
        else:
            range_name = "from current location to end of file"
            replace_range = IndexRange(maintext().get_insert_index(), maintext().end())
    return replace_range, range_name


def message_from_regex_exception(exc: re.error) -> str:
    """Create error message from regex exception.

    Args:
        exc - The regex exception to describe.
    """
    message = str(exc)
    message = message[0].upper() + message[1:]
    return message + " in regex " + exc.pattern  # type:ignore[attr-defined]



"""Store and manipulate page labels."""

import tkinter as tk
from tkinter import simpledialog, ttk
from typing import Optional, Any

import roman  # type: ignore[import-untyped]

from guiguts.maintext import maintext, page_mark_from_img
from guiguts.utilities import is_mac
from guiguts.widgets import OkApplyCancelDialog, mouse_bind, ToolTip

STYLE_COLUMN = "#2"
STYLE_ARABIC = "Arabic"
STYLE_ROMAN = "Roman"
STYLE_DITTO = '"'
NUMBER_COLUMN = "#3"
NUMBER_PAGENUM = "Number"
NUMBER_INCREMENT = "+1"
NUMBER_NONE = "No Count"
COL_HEAD_IMG = "Img"
COL_HEAD_STYLE = "Style"
COL_HEAD_NUMBER = "Number"
COL_HEAD_LABEL = "Label"
PAGE_LABEL_PREFIX = "Pg "

STYLES = [STYLE_ARABIC, STYLE_ROMAN, STYLE_DITTO]
NUMBERS = [NUMBER_PAGENUM, NUMBER_INCREMENT, NUMBER_NONE]


class PageDetail(dict):
    """Page detail information for one page.

    Attributes:
        index: Index of start of page in file.
        style: Page number style: "Arabic" or "Roman"
        number:
    """

    def __init__(
        self, index: str, style: str = '"', number: str = NUMBER_INCREMENT
    ) -> None:
        """Initialize a PageDetail object to hold page information."""
        self["index"] = index
        self["style"] = style
        self["number"] = number
        self["label"] = ""


class PageDetails(dict[str, PageDetail]):
    """Dictionary of Page Detail objects: key is png name."""

    def recalculate(self) -> None:
        """Recalculate labels from details."""
        number = 0
        style = STYLE_ARABIC
        for _, detail in sorted(self.items()):
            if detail["style"] != STYLE_DITTO:
                style = detail["style"]
            if detail["number"] == NUMBER_NONE:
                detail["label"] = ""
            else:
                if detail["number"] == NUMBER_INCREMENT:
                    number += 1
                else:
                    number = int(detail["number"])
                if style == STYLE_ROMAN:
                    try:
                        label = roman.toRoman(number).lower()
                    except roman.OutOfRangeError:
                        label = "Roman error"
                else:
                    label = str(number)
                detail["label"] = PAGE_LABEL_PREFIX + label

    def png_from_label(self, label: str) -> str:
        """Find the png corresponding to the page with the given label.

        Args:
            label: The label to match against.

        Returns:
            The png number corresponding to the given label, or empty string.
        """
        for png, detail in self.items():
            if detail["label"] == label:
                return png
        return ""

    def copy_details_from(self, details: "PageDetails") -> None:
        """Copy the style and number details from given PageDetails

        Args:
            details: Details to copy from.
        """
        for img, detail in details.items():
            self[img] = PageDetail("", detail["style"], detail["number"])
        self.recalculate()


class PageDetailsDialog(OkApplyCancelDialog):
    """A dialog that allows the user to view/edit page details.

    Attributes:
        master_details: Dictionary of page details as used by text file.
        details: Temporary dictionary to hold values as changed in dialog.
        changed: True if any changes have been made in dialog.
    """

    manual_page = "File_Menu#Configure_Page_Markers_and_Page_Labels"

    def __init__(self, page_details: PageDetails) -> None:
        """Initialize class members from page details."""
        super().__init__("Configure Page Labels")
        self.master_details = page_details
        self.details = PageDetails()
        self.details.copy_details_from(self.master_details)
        self.changed = False

        columns = (COL_HEAD_IMG, COL_HEAD_STYLE, COL_HEAD_NUMBER, COL_HEAD_LABEL)
        widths = (50, 80, 80, 120)
        self.list = ttk.Treeview(
            self.top_frame,
            columns=columns,
            show="headings",
            height=10,
            selectmode=tk.BROWSE,
        )
        ToolTip(
            self.list,
            "\n".join(
                [
                    "Click in style column to cycle Arabic/Roman/Ditto",
                    "Click in number column to cycle +1/No Count/Set Number",
                    "Shift click to cycle in reverse order",
                ]
            ),
            use_pointer_pos=True,
        )
        for col, column in enumerate(columns):
            self.list.column(
                f"#{col + 1}",
                minwidth=10,
                width=widths[col],
                stretch=False,
                anchor=tk.CENTER,
            )
            self.list.heading(f"#{col + 1}", text=column)

        mouse_bind(
            self.list, "1", lambda event: self.item_clicked(event, reverse=False)
        )
        mouse_bind(
            self.list, "Shift+1", lambda event: self.item_clicked(event, reverse=True)
        )

        def display_page(_event: Optional[tk.Event] = None) -> None:
            """Display the page for the selected row."""
            try:
                png = self.list.set(self.list.focus())[COL_HEAD_IMG]
            except KeyError:
                return
            index = maintext().rowcol(page_mark_from_img(png))
            maintext().set_insert_index(index, focus=False)

        def select_by_index(idx: int) -> None:
            """Select the item with the given index (-1 for last one)."""
            try:
                child = self.list.get_children()[idx]
            except IndexError:
                return
            self.select_and_focus(child)  # S
            self.list.see(child)
            display_page()

        self.bind("<Home>", lambda _e: select_by_index(0))
        self.bind("<End>", lambda _e: select_by_index(-1))
        if is_mac():
            self.bind("<Command-Up>", lambda _e: select_by_index(0))
            self.bind(
                "<Command-Down>",
                lambda _e: select_by_index(-1),
            )

        self.list.bind("<<TreeviewSelect>>", display_page)
        self.list.grid(row=0, column=0, sticky=tk.NSEW)

        self.scrollbar = ttk.Scrollbar(
            self.top_frame, orient=tk.VERTICAL, command=self.list.yview
        )
        self.list.configure(yscroll=self.scrollbar.set)  # type: ignore[call-overload]
        self.scrollbar.grid(row=0, column=1, sticky=tk.NS)

        self.populate_list(self.details)

        # Position list at current page
        if cur_img := maintext().get_current_image_name():
            children = self.list.get_children()
            for idx, child in enumerate(children):
                png = self.list.set(child)[COL_HEAD_IMG]
                if png == cur_img:
                    self.select_and_focus(child)
                    # "see" puts item at top, so see the position a few earlier
                    self.list.see(children[max(0, idx - 3)])
                    break

        self.list.focus_set()

    def select_and_focus(self, item: str) -> None:
        """Select and set focus to the given item.

        Although for the "browse" select mode used in this TreeView, there is
        only one selected item, there can be several items selected in other
        TreeViews, and one item can have focus. Since user can use mouse & arrow
        keys to change selection, and we also change it programmatically to
        support use of Home/End keys, it's necessary to ensure the selection and
        the focus always point to the same item. Otherwise using Home to go to
        top of list then arrow key to move to second item will not work.

        Args:
            item: The item to be selected/focused.
        """
        self.list.selection_set(item)
        self.list.focus(item)

    def populate_list(self, details: PageDetails, see_index: int = 0) -> None:
        """Populate the page details list from the given details.

        Args:
            details: PageDetails to be displayed in dialog.
            see_index: Index of item in list that should be made visible.
        """
        children = self.list.get_children()
        for child in children:
            self.list.delete(child)

        for png, detail in sorted(details.items()):
            entry = (png, detail["style"], detail["number"], detail["label"])
            self.list.insert("", tk.END, values=entry)

        children = self.list.get_children()
        if children:
            self.select_and_focus(children[see_index])
            self.list.see(children[see_index])

    def item_clicked(self, event: tk.Event, reverse: bool) -> None:
        """Called when page detail item is clicked.

        If click is in style or number column, then advance style/number
        setting to the next value. Refresh the list to show new labels.

        Args:
            event: Event containing location of mouse click
            reverse: True to "advance" in reverse!
        """
        row_id = self.list.identify_row(event.y)
        row = self.list.set(row_id)

        col_id = self.list.identify_column(event.x)
        if col_id not in (STYLE_COLUMN, NUMBER_COLUMN):
            return

        if col_id == STYLE_COLUMN:
            if COL_HEAD_STYLE not in row:
                return
            # Click in style column advances/retreats style
            style_index = STYLES.index(row[COL_HEAD_STYLE])
            if reverse:
                style_index = len(STYLES) - 1 if style_index == 0 else style_index - 1
            else:
                style_index = 0 if style_index == len(STYLES) - 1 else style_index + 1
            self.details[row[COL_HEAD_IMG]]["style"] = STYLES[style_index]
        elif col_id == NUMBER_COLUMN:
            if COL_HEAD_NUMBER not in row:
                return
            # Click in number column advances/retreats number type.
            # May need to prompt user for page number.
            value = row[COL_HEAD_NUMBER]
            if value.isdecimal():
                value = NUMBER_PAGENUM
            number_index = NUMBERS.index(value)
            if reverse:
                number_index = (
                    len(NUMBERS) - 1 if number_index == 0 else number_index - 1
                )
            else:
                number_index = (
                    0 if number_index == len(NUMBERS) - 1 else number_index + 1
                )
            new_value: int | Any = NUMBERS[number_index]
            if new_value == NUMBER_PAGENUM:
                pagenum = simpledialog.askinteger(
                    "Set Number", "Enter page number", parent=self
                )
                new_value = pagenum if pagenum else value
            self.details[row[COL_HEAD_IMG]]["number"] = str(new_value)

        # Refresh the list
        self.details.recalculate()
        self.populate_list(self.details, self.list.index(row_id))
        self.changed = True

    def apply_changes(self) -> bool:
        """Overridden to update page label settings from the dialog."""
        if self.changed:
            self.master_details.copy_details_from(self.details)
            maintext().set_modified(True)
        return True



"""Levenshtein edit distance check tool"""

from enum import IntEnum, auto
import importlib.resources
import logging
import time
from typing import Any
from tkinter import ttk

from Levenshtein import distance
import regex as re

from guiguts.checkers import CheckerDialog
from guiguts.data import dictionaries
from guiguts.file import ProjectDict
from guiguts.maintext import maintext
from guiguts.misc_tools import tool_save
from guiguts.preferences import PersistentInt, PrefKey, preferences
from guiguts.spell import get_spell_checker
from guiguts.utilities import IndexRowCol, IndexRange

logger = logging.getLogger(__package__)

DEFAULT_DICTIONARY_DIR = importlib.resources.files(dictionaries)

SPELL_CHECK_OK_YES = 0
SPELL_CHECK_OK_NO = 1
SPELL_CHECK_OK_BAD = 2

LINES_IN_REPORT_LIMIT = 4

_the_levenshtein_checker = None  # pylint: disable=invalid-name

#########################################################
# levenshtein.py
# Python author: Quentin Campbell (DP:qgc) - 2024.
# Based on the processing logic in the PPtxt Go version
# authored by Roger Franks (DP:rfrank - 2020) but using
# a publicly available Levenshtein module authored by
# Max Bachmann.
#########################################################


class LevenshteinEditDistance(IntEnum):
    """Enum class to store Levenshtein Edit Distance."""

    ONE = auto()
    TWO = auto()


class LevenshteinCheckerDialog(CheckerDialog):
    """Levenshtein Checker Dialog."""

    manual_page = "Tools_Menu#Word_Distance_Check"

    def __init__(self, **kwargs: Any) -> None:
        """Initialize Levenshtein checker dialog."""
        super().__init__(
            "Levenshtein Edit Distance Check",
            tooltip="\n".join(
                [
                    "Left click: Select & find highlighted word in file",
                    "Right click: Remove line with highlighted word from list",
                    "Shift Right click: Also remove all matching lines from list",
                ]
            ),
            **kwargs,
        )

        self.edit_distance = PersistentInt(PrefKey.LEVENSHTEIN_DISTANCE)

        frame = ttk.Frame(self.header_frame)
        frame.grid(column=0, row=1, sticky="NSEW")
        ttk.Label(
            frame,
            text="Edit Distance:",
        ).grid(row=0, column=1, sticky="NSE", padx=(0, 5))
        ttk.Radiobutton(
            frame,
            text="1",
            variable=self.edit_distance,
            value=LevenshteinEditDistance.ONE,
            takefocus=False,
        ).grid(row=0, column=2, sticky="NSE", padx=2)
        ttk.Radiobutton(
            frame,
            text="2",
            variable=self.edit_distance,
            value=LevenshteinEditDistance.TWO,
            takefocus=False,
        ).grid(row=0, column=3, sticky="NSE", padx=2)


def run_levenshtein_check_on_file(project_dict: ProjectDict) -> None:
    """Check Levenshtein edit distance between selected file words."""

    ####
    # The following are declared and accessed as 'nonlocal' variables
    # by functions defined within the enclosing scope of this top-most
    # function definition. The names must exist when first encountered
    # as they cannot be created by a first assignment in a nested def.
    ####

    # Each word that is correctly spell-checked is appended to this list,
    # hence can contain many duplicates of a word in all its case forms.
    # E.g. ['another', 'Another', 'another', ...]
    good_words: list[str] = []
    # Each word that fails the spell check is appended to this list, hence
    # can contain many duplicates of a word in all its case forms.
    # E.g. ['anither', 'Anither', 'anither', ...]
    suspect_words: list[str] = []
    # Key is a word in whatever case it appears in the text. Maps
    # the frequency with which that word form appears in the text.
    all_words_counts: dict[str, int] = {}
    # Key is a word in whatever case it appears in the text. Maps
    # the word to the line(s) it appears on using a list of
    # tuples: (line_number, start_columan_index).
    # E.g. 'another': [(1, 8), (3, 0), (3, 12)]
    word_to_lines_map: dict[str, list[tuple]] = {}
    # A list of tuples, each tuple containing two, lowercase, words.
    # These are the 'suspect_word/good_word' pairs that differ by the
    # Levenshtein edit distance specified (1 or 2). E.g. of content:
    # [('anither', 'another'), ...]
    distance_check_results: list[tuple] = []
    # Given a lowercase word as a key, the value of that key is a list
    # of different case versions of the key that appear in the text.
    # E.g. 'another': ['Another', 'another', 'ANOTHER']
    all_words_case_map: dict[str, list[str]] = {}

    def build_header_line(result_tuple: tuple) -> str:
        """Function to build a header line for display

        It processes a tuple that contains a pair of words that
        are the required Levenshtein edit distance apart. E.g.
        ('anither', 'another')

        Args:
            result_tuple: contains a pair of lowercase words.

        Returns:
            Formatted string. E.g.
                Anither (1), anither (2) <-> ANOTHER (1), Another (3), another (7)
        """

        # The tuple argument contains two lowercase strings.
        suspect_word = result_tuple[0]
        test_word = result_tuple[1]

        # Find any different-case versions of each word. The key
        # for this map is always the canonical form of the word;
        # that is, the lowercase version.
        #
        # Note that the strings in an 'all_words_case_map[word]' list
        # are not necessarily in lexicographic order so always sort it.
        # E.g.:
        #     ['Another', 'another', 'ANOTHER']
        # when we want to display
        #     ['ANOTHER', 'Another', 'another'']
        list_sorted_sw = sorted(all_words_case_map[suspect_word])
        list_sorted_gw = sorted(all_words_case_map[test_word])

        # When the same word appears in different cases build a header
        # line that reflects that. E.g:
        #
        #   Anither(1), anither(2) <-> ANOTHER(1), Another(2), another(5)
        #
        # Build it in two sections, left and right parts with the
        # separator between them.

        # NB In the following section the key to the 'all_words_counts'
        # dictionary can be in any case form; that is, they are not
        # expected to be in lowercase canonical form.

        # This half displays one or more case versions of the suspect word.
        left_part = ""
        for word in list_sorted_sw:
            left_part += f"{word}({all_words_counts[word]}), "
        # Chop the trailing ", " characters.
        left_part = left_part[:-2]

        # This half displays one or more case versions of the test word.
        right_part = ""
        for word in list_sorted_gw:
            right_part += f"{word}({all_words_counts[word]}), "
        # Chop the trailing ", " characters.
        right_part = right_part[:-2]

        return f"{left_part} <-> {right_part}"

    def sort_tuples(tuples: list[tuple], key_indx: int) -> list[tuple]:
        """Utility function to sort a list of tuples

        Args:
            tuples: a list of two-element tuples
            key_indx: int 0 or 1 - element of a tuple on which to sort

        Returns:
            Input list of tuples sorted on key_indx element of each tuple
        """

        # Define a key function that returns the desired element of each tuple.
        def key_func(x: tuple) -> int:
            return x[key_indx]

        # key_func = lambda x: x[key_indx]

        # Use the 'sorted' function to sort the list of tuples using the key function.
        sorted_tuples = sorted(tuples, key=key_func)

        return sorted_tuples

    def get_sorted_list_of_linecol_tuples(word: str) -> list:
        """Utility function to sort line_number/column_start_index tuples

        Enables the reporting of every line on which a word occurs (in any of its
        case forms) by increasing order of line number and column index.

        Arg:
            word: all line_number/column_start_index tuples for this word will be
                    collected and sorted.

        Returns:
            list of sorted tuples.
        """

        # The 'word' argument is in canonical form. Put the different
        # case versions of the word into 'word_list'.
        word_list = all_words_case_map[word]

        # Each case version of the word is mapped to a list of tuples.
        # Each tuple represents (line_number, column_start_index) as
        # integers. The list of tuples for each case version represents
        # the location of every occurrence of that form of the word in
        # the file.
        #
        # Gather together all the tuples of each different case version
        # of a word and sort that list into start column order within line
        # number order.
        tuples = []

        for word_entry in word_list:
            # The key to the map below is in any case form.
            tuples_list = word_to_lines_map[word_entry]
            for tuple_entry in tuples_list:
                tuples.append(tuple_entry)

        # First sort the tuples by column_start_index
        sorted_tuples = sort_tuples(tuples, 1)
        # Then sort the result by line_number
        sorted_tuples = sort_tuples(sorted_tuples, 0)
        # If the original list of tuples was, say
        #   [(2, 5), (1, 2), (4, 4), (2, 3)]
        # it should now be
        #   [(1, 2), (2, 3), (2, 5), (4, 4)]

        return sorted_tuples

    def unpack_and_write_to_dialog(result_tuple: tuple) -> None:
        """Function for reporting word pair results from distance check

        This function is called once for each pair of words that meet
        the required Levenshtein edit distance criteria. It calls utility
        functions to generate the header line and select and highlight all
        the file lines that make up the report for that word pair. It then
        calls checker dialog methods to report the header line and all the
        accompanying file lines.

        Arg:
            result_tuple: 2-element tuple containing a lowercase 'suspect'
                            word and a lowercase 'good' word against which
                            it was tested.
        """

        # The tuple argument contains two lowercase strings; that is,
        # they are the canonical form of each word of the pair being
        # reported.
        suspect_word = result_tuple[0]
        test_word = result_tuple[1]

        # In order to build the header line for the report; e.g.
        #   Anither(1), anither(2) <-> ANOTHER(1), Another(2), another(5)
        # will need all other case forms (if any) of the two words. Add
        # the generated header line to the report.
        checker_dialog.add_header(build_header_line(result_tuple))

        # The header line is first followed by lines that contain the 'suspect'
        # word in all of it case forms. The number of lines output is limited
        # to LINES_IN_REPORT_LIMIT, currently set to 4. If there were more lines
        # to output to the dialog then the lines that are displayed are followed
        # by a line displaying "...more". The occurrence of the word on each
        # line is highlighted in the usual way.
        tuples = get_sorted_list_of_linecol_tuples(suspect_word)
        lines_output_count = 0
        for index_tuple in tuples:
            # Limit the number of 'suspect' word lines reported.
            if lines_output_count == LINES_IN_REPORT_LIMIT:
                lines_output_count = -1
                break
            error_start, error_end = make_into_strings(index_tuple, len(suspect_word))
            start_rowcol = IndexRowCol(error_start)
            end_rowcol = IndexRowCol(error_end)
            file_line = maintext().get(
                error_start + " linestart", error_end + " lineend"
            )
            checker_dialog.add_entry(
                file_line,
                IndexRange(start_rowcol, end_rowcol),
                start_rowcol.col,
                end_rowcol.col,
            )
            lines_output_count += 1

        if lines_output_count < 0:
            checker_dialog.add_footer("  ...more")

        # No more lines to output for the suspect word so add a separator.
        checker_dialog.add_header("----")

        # The separator line is then followed by lines that contain the 'good'
        # word in all of it case forms. The number of lines output is limited
        # to LINES_IN_REPORT_LIMIT, currently set to 4. If there were more lines
        # to output to the dialog then the lines that are displayed are followed
        # by a line displaying "...more". The occurrence of the word on each
        # line is highlighted in the usual way.
        tuples = get_sorted_list_of_linecol_tuples(test_word)
        lines_output_count = 0
        for index_tuple in tuples:
            # Limit the number of 'good' word lines reported.
            if lines_output_count == LINES_IN_REPORT_LIMIT:
                lines_output_count = -1
                break
            error_start, error_end = make_into_strings(index_tuple, len(test_word))
            start_rowcol = IndexRowCol(error_start)
            end_rowcol = IndexRowCol(error_end)
            file_line = maintext().get(
                error_start + " linestart", error_end + " lineend"
            )
            checker_dialog.add_entry(
                file_line,
                IndexRange(start_rowcol, end_rowcol),
                start_rowcol.col,
                end_rowcol.col,
            )
            lines_output_count += 1

        if lines_output_count < 0:
            checker_dialog.add_footer("  ...more")

        # Add a blank line to separate this word-pair report from that of the
        # next pair.
        checker_dialog.add_footer("")

    def make_into_strings(index_tuple: tuple, hilite_length: int) -> tuple[str, str]:
        """Utility function to generate error_start & error_end strings from int values

        Args:
            index_tuple: an integer line number and an integer column start index
            hilite_length: integer length of string to be lighlighted on the line.

        Return:
            An error_start string (E.g. "18.3" and an error_end string (E.g. "18.9")
        """

        line_num = str(index_tuple[0])
        col_indx = str(index_tuple[1])
        error_start = f"{line_num}.{col_indx}"
        col_indx = str(index_tuple[1] + hilite_length)
        error_end = f"{line_num}.{col_indx}"

        return error_start, error_end

    def reject_this(word: str) -> bool:
        """Helper function that extracts repeated code. Decides whether
        a word should be added to a list ('good' or 'suspects') and thus
        will be distance checked. Criteria for rejection includes being
        too short, a Roman numeral, all digits, etc.

        Args:
            word: should we add this word to a list or not?

        Return:
            A boolean. True if answer is 'No'; i.e. it won't be distance checked
        """

        # Reject single-letter 'words' as they generate spurious report lines
        if len(word) == 1:
            return True
        # Reject if word contains Greek letters; e.g. 5μ, 3π, etc.
        if re.search(r"\p{Greek}+", word):
            return True
        # Reject word if a valid Roman numeral
        worduc = word.upper()
        if re.fullmatch(
            r"(?=[MDCLXVI])M*(C[MD]|D?C{0,3})(X[CL]|L?X{0,3})(I[XV]|V?I{0,3})",
            worduc,
        ):
            return True
        # Otherwise the length/type of the word looks OK. Keep it so that it's
        # added to one of the two lists.
        return False

    def map_words_in_file(project_dict: ProjectDict) -> None:
        """Spell check the currently loaded file and extract a list of suspect
        words and a list of good words. Build a map of the line position index(es)
        of each word and a map of the frequency of each word. In addition, build
        a map of the case variants of a word that uses the lowercase form of the
        word as key.

        The two lists of words from which word pairs are generated for possible
        distance checking are unique collections of words all in lowercase. This
        is crucial to minimise execution time needed to generate the reported
        results.

        However when we generate the edit distance report we want to display all
        the case variants of a word (e.g. 'ANOTHER', 'Another', 'another') hence
        the role of the 'all_words_case_map' created in this function.

        Outputs into a nonlocal list or dictionary:
            List of suspect words - each an apparent spelling error.
            Counts of those words (dictionary).
            List of good words - each found in an English dictionary.
            Counts of those words (dictionary).
            Map of the line index(es) of each word and hilite start index on line
                (dictionary).
            Map of the case variants of a word (dictionary).
        """
        assert _the_spell_checker is not None

        # The following must already exist (see enclosing function) before they are
        # used here. They are the lists and dictionaries into which the results of
        # this function are 'returned'.
        nonlocal good_words
        nonlocal suspect_words
        nonlocal word_to_lines_map, all_words_counts
        nonlocal all_words_case_map

        # Do spelling checks below using SpellChecker methods from spell.py

        # The main loop from spell.py with some minor adaptions. It splits each
        # file line into 'words' and calls SpellChecker.spell_check_word() on each.
        # Depending on the result, the word is appended either to 'good_words' list
        # (they are in the spelling dictionary) or 'suspect_words' list (not in the
        # spelling dictionary). Some variable names changed to avoid pylint flagging
        # 'duplicate code' with the same block of code in spell.py.
        for line_text, line_number in maintext().get_lines():
            words = re.split(r"[^\p{Alnum}\p{Mark}'’]", line_text)
            next_column = 0
            for word in words:
                column = next_column
                next_column = column + len(word) + 1
                if word:
                    # Consider whether 'word' should be added to the 'good' or the
                    # 'suspects' list. It might not be added to either if it is
                    # too short or a Roman numeral, etc., so is unsuitable to be
                    # distance checked.
                    spell_check_result = _the_spell_checker.spell_check_word(
                        word, project_dict
                    )
                    if spell_check_result == SPELL_CHECK_OK_YES:
                        # Don't add it to the list if unsuitable for distance checking.
                        if reject_this(word):
                            continue
                        good_words.append(word)
                        try:
                            all_words_counts[word] += 1
                        except KeyError:
                            all_words_counts[word] = 1

                        try:
                            word_to_lines_map[word].append((line_number, column))
                        except KeyError:
                            word_to_lines_map[word] = [(line_number, column)]
                        continue

                    # If word has leading straight apostrophe, it might be
                    # open single quote; trim it and check again
                    if spell_check_result == SPELL_CHECK_OK_NO and word.startswith("'"):
                        word = word[1:]
                        spell_check_result = _the_spell_checker.spell_check_word(
                            word, project_dict
                        )
                        if spell_check_result == SPELL_CHECK_OK_YES:
                            # Don't add it to the list if unsuitable for distance checking.
                            if reject_this(word):
                                continue
                            good_words.append(word)
                            try:
                                all_words_counts[word] += 1
                            except KeyError:
                                all_words_counts[word] = 1

                            try:
                                word_to_lines_map[word].append((line_number, column))
                            except KeyError:
                                word_to_lines_map[word] = [(line_number, column)]
                            continue

                    # If trailing straight/curly apostrophe, it might be
                    # close single quote; trim it and check again
                    if spell_check_result == SPELL_CHECK_OK_NO and re.search(
                        r"['’]$", word
                    ):
                        word = word[:-1]
                        spell_check_result = _the_spell_checker.spell_check_word(
                            word, project_dict
                        )
                        if spell_check_result == SPELL_CHECK_OK_YES:
                            # Don't add it to the list if unsuitable for distance checking
                            if reject_this(word):
                                continue
                            good_words.append(word)
                            try:
                                all_words_counts[word] += 1
                            except KeyError:
                                all_words_counts[word] = 1

                            try:
                                word_to_lines_map[word].append((line_number, column))
                            except KeyError:
                                word_to_lines_map[word] = [(line_number, column)]
                            continue

                    # Word not in dictionary. If word is not now empty and not
                    # consisting only of single quotes, add it to suspect words
                    # list.
                    if not re.fullmatch(r"['’]*", word):
                        # Don't add it to the list if unsuitable for distance checking
                        if reject_this(word):
                            continue
                        suspect_words.append(word)
                        try:
                            all_words_counts[word] += 1
                        except KeyError:
                            all_words_counts[word] = 1

                        try:
                            word_to_lines_map[word].append((line_number, column))
                        except KeyError:
                            word_to_lines_map[word] = [(line_number, column)]

        # NB The lists of 'good' words and 'suspect' words are not always mutually
        # exclusive since a lowercase version of a word may not be in the spelling
        # dictionary used but the uppercase version of the word is in that dictionary.
        # An example of this is 'kirk' which is not in spelling dictionary but 'Kirk'
        # is in the dictionary.

        # Build a dictionary, keyed by the lower case version of a word, whose value
        # will be a list of the same word in the different cases in which it appears
        # in the file. E.g.
        #  dict['another'] -> ['ANOTHER', Another', 'another']
        #
        # NB Only words that appear in the text will appear in that list. Thus the
        #    lowercase version of a word that's used as a key will not be in the list
        #    unless it also appears in the text.

        # As stated in the comments above, 'word' may be a string in any case.
        for word in good_words:
            # The lowercase version of the word is the dictionary key.
            word_lc = word.lower()
            if word_lc in all_words_case_map:
                # There's a dictionary entry keyed by the lowercase form of the word.
                # Is 'word' in the entries list? That is, if key is 'another' it is
                # asking here whether, say, 'Another' is in the values list for the
                # key.
                if word not in all_words_case_map[word_lc]:
                    # Then add it to the list
                    all_words_case_map[word_lc].append(word)
            else:
                # Create the dictionary entry which is a list.
                all_words_case_map[word_lc] = [word]

        # Same comments apply here.
        for word in suspect_words:
            word_lc = word.lower()
            if word_lc in all_words_case_map:
                # There's a dictionary entry. Is 'word' in the entries list?
                if word not in all_words_case_map[word_lc]:
                    # Then add it to the list
                    all_words_case_map[word_lc].append(word)
            else:
                # Create the dictionary entry which is a list.
                all_words_case_map[word_lc] = [word]

    def distance_check_words() -> None:
        """Function to select, and edit distance check, pairs of suspect/good words.

        The process involves taking each lowercase form of a suspect word and doing a
        Levenshtein edit distance check against the lowercase form of every good word.

        Keeping the lengths of the respective lists of words as short as possible is
        critical. A book quoting a lot of speech in dialect may have a 'suspect' words
        list of more than 1,000 entries. If the 'good' words list has 8,000 or 9,000
        entries then the number of pairs of suspect<->good words to consider will approach
        10 million. Processing that number of pairs is very demanding of CPU time.
        """

        nonlocal distance_check_results

        # As a word may appear many times in a text, and in many case forms, for example
        # 'ANOTHER', 'Another', 'another, each of the two lists is kept as short as
        # possible by containing just a single lowercase instance of the word: 'another'
        # in the example above.

        # NB For display purposes a separate dictionary maps each case form of a word to
        # its line number and column start index for every occurrence in the file. That is
        # in turn indexed by entries from another dictionary that maps a word in lowercase
        # to any other case form that appears in the file. E.g.: 'another' -> 'ANOTHER',
        # 'Another', 'another'.
        #
        # These dictionaries are only used when processing for display in the report the
        # small number of word pairs that meet the required Levenshtein edit distance criteria.
        # Crucially they free us to do the CPU-intensive word selection and Levenshtein edit
        # distance calculations on the smallest feasible set of word pairs.

        # First make every word lowercase in a copy of each list.
        suspect_words_unique = [x.lower() for x in suspect_words]
        good_words_unique = [x.lower() for x in good_words]
        # The resulting lists still contain multiple occurrences of a word. Use set() to
        # create a unique collection of words from each list. Then turn the result back
        # into a list and sort it.
        suspect_words_unique = sorted(list(set(suspect_words_unique)))
        good_words_unique = sorted(list(set(good_words_unique)))

        # For each lowercase form of a suspect word do distance check against the lowercase
        # form of every good word; i.e. 'test' words. Some obvious filtering of both suspect
        # and test words is done to eliminate some pairs before the edit distance calculation
        # check is made.
        for suspect_wordlc in suspect_words_unique:
            # Must be five letters or more OR contain an unexpected character.
            unexpected = re.sub(r"[a-z0-9’'æœ]", "", suspect_wordlc)
            if len(suspect_wordlc) < 3 and len(unexpected) == 0:
                continue
            for test_wordlc in good_words_unique:
                # No point in processing a suspect<->test word pair whose difference in
                # length is greater than the edit distance to be used in the check (1 or 2).
                if abs(len(suspect_wordlc) - len(test_wordlc)) > distance_to_check:
                    continue
                # Check if they are the same word differing only by capitalisation.
                # This check may seem unnecessary since the two lists are supposed
                # to contain mutually exclusive collections of words. This may not
                # always be so. For example 'kirk', the Scottish word for a church
                # is not in the dictionary so would appear in 'suspect_words' while
                # the name 'Kirk' IS in the dictionary hence would appear in the
                # 'good_words/test_words' list.
                if suspect_wordlc == test_wordlc:
                    continue

                # Calculate Levenshtein distance
                dist = distance(
                    suspect_wordlc, test_wordlc, score_cutoff=distance_to_check
                )

                if dist == distance_to_check:
                    distance_check_results.append((suspect_wordlc, test_wordlc))

    ####
    # Executable section of enclosing function 'run_levenshtein_check_on_file'.
    ####

    # Start time of prgram execution
    prog_start = time.time()

    # Set up SpellChecker here before the tool window is opened in
    # case it cannot find the required dictionaries.
    _the_spell_checker = get_spell_checker()
    if _the_spell_checker is None:
        return

    # Create the checker dialog to show results
    checker_dialog = LevenshteinCheckerDialog.show_dialog(
        rerun_command=lambda: levenshtein_check(project_dict),
    )

    # Get the edit distance used last time Levenshtein was run or default if first run.
    distance_to_check = preferences.get(PrefKey.LEVENSHTEIN_DISTANCE)

    # Build maps of the good words and the suspect words in the file.
    # Ideally should only do this once and not every time the tool is
    # rerun.
    map_words_in_file(project_dict)

    # For each suspect word, check against each good word in turn.
    # Very compute intensive. Where 90% of the execution time is consumed.
    distance_check_words()

    # Did distance_check_words() find any word pairs that met
    # the required Levenshtein criteria?
    if len(distance_check_results) == 0:
        # No.
        checker_dialog.add_entry("No Levenshtein edit distance queries reported")
        checker_dialog.add_entry("")
    else:
        # Yes, so display the results using checker_dialog methods.
        for result_tuple in distance_check_results:
            unpack_and_write_to_dialog(result_tuple)

    # Calculate and display execution time of run.
    prog_end = time.time()
    checker_dialog.add_footer(f"Execution time: {(prog_end - prog_start):.2f} seconds")

    checker_dialog.display_entries()


def levenshtein_check(project_dict: ProjectDict) -> None:
    """Do Levenshtein edit distance checks"""

    if not tool_save():
        return

    run_levenshtein_check_on_file(project_dict)
